@@ -1,42 +1,53 @@
 import datetime
+import pathlib
+import pickle
 import re
-import time
 from collections import defaultdict
 from collections.abc import Mapping
-from http.cookies import Morsel, SimpleCookie
+from http.cookies import Morsel
 from math import ceil
-from urllib.parse import urlsplit
+from yarl import URL
 
 from .abc import AbstractCookieJar
-from .helpers import is_ip_address
+from .helpers import SimpleCookie, is_ip_address
 
 
 class CookieJar(AbstractCookieJar):
     """Implements cookie storage adhering to RFC 6265."""
 
     DATE_TOKENS_RE = re.compile(
-        "[\x09\x20-\x2F\x3B-\x40\x5B-\x60\x7B-\x7E]*"
-        "(?P<token>[\x00-\x08\x0A-\x1F\d:a-zA-Z\x7F-\xFF]+)")
+        r"[\x09\x20-\x2F\x3B-\x40\x5B-\x60\x7B-\x7E]*"
+        r"(?P<token>[\x00-\x08\x0A-\x1F\d:a-zA-Z\x7F-\xFF]+)")
 
-    DATE_HMS_TIME_RE = re.compile("(\d{1,2}):(\d{1,2}):(\d{1,2})")
+    DATE_HMS_TIME_RE = re.compile(r"(\d{1,2}):(\d{1,2}):(\d{1,2})")
 
-    DATE_DAY_OF_MONTH_RE = re.compile("(\d{1,2})")
+    DATE_DAY_OF_MONTH_RE = re.compile(r"(\d{1,2})")
 
-    DATE_MONTH_RE = re.compile(
-        "(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)", re.I)
+    DATE_MONTH_RE = re.compile("(jan)|(feb)|(mar)|(apr)|(may)|(jun)|(jul)|"
+                               "(aug)|(sep)|(oct)|(nov)|(dec)", re.I)
 
-    DATE_YEAR_RE = re.compile("(\d{2,4})")
+    DATE_YEAR_RE = re.compile(r"(\d{2,4})")
 
-    MAX_TIME = time.mktime((2100, 1, 1, 1, 1, 1, 1, 1, 1,))  # so far in future
+    MAX_TIME = 2051215261.0  # so far in future (2035-01-01)
 
     def __init__(self, *, unsafe=False, loop=None):
         super().__init__(loop=loop)
         self._cookies = defaultdict(SimpleCookie)
         self._host_only_cookies = set()
         self._unsafe = unsafe
-        self._next_expiration = ceil(loop.time())
+        self._next_expiration = ceil(self._loop.time())
         self._expirations = {}
 
+    def save(self, file_path):
+        file_path = pathlib.Path(file_path)
+        with file_path.open(mode='wb') as f:
+            pickle.dump(self._cookies, f, pickle.HIGHEST_PROTOCOL)
+
+    def load(self, file_path):
+        file_path = pathlib.Path(file_path)
+        with file_path.open(mode='rb') as f:
+            self._cookies = pickle.load(f)
+
     def clear(self):
         self._cookies.clear()
         self._host_only_cookies.clear()
@@ -77,10 +88,9 @@ class CookieJar(AbstractCookieJar):
         self._next_expiration = min(self._next_expiration, when)
         self._expirations[(domain, name)] = when
 
-    def update_cookies(self, cookies, response_url=None):
+    def update_cookies(self, cookies, response_url=URL()):
         """Update cookies."""
-        url_parsed = urlsplit(response_url or "")
-        hostname = url_parsed.hostname
+        hostname = response_url.raw_host
 
         if not self._unsafe and is_ip_address(hostname):
             # Don't accept cookies from IPs
@@ -120,7 +130,7 @@ class CookieJar(AbstractCookieJar):
             path = cookie["path"]
             if not path or not path.startswith("/"):
                 # Set the cookie's path to the response path
-                path = url_parsed.path
+                path = response_url.path
                 if not path.startswith("/"):
                     path = "/"
                 else:
@@ -153,13 +163,12 @@ class CookieJar(AbstractCookieJar):
 
         self._do_expiration()
 
-    def filter_cookies(self, request_url):
+    def filter_cookies(self, request_url=URL()):
         """Returns this jar's cookies filtered by their attributes."""
         self._do_expiration()
-        url_parsed = urlsplit(request_url)
         filtered = SimpleCookie()
-        hostname = url_parsed.hostname or ""
-        is_not_secure = url_parsed.scheme not in ("https", "wss")
+        hostname = request_url.raw_host or ""
+        is_not_secure = request_url.scheme not in ("https", "wss")
 
         for cookie in self:
             name = cookie.key
@@ -179,13 +188,17 @@ class CookieJar(AbstractCookieJar):
             elif not self._is_domain_match(domain, hostname):
                 continue
 
-            if not self._is_path_match(url_parsed.path, cookie["path"]):
+            if not self._is_path_match(request_url.path, cookie["path"]):
                 continue
 
             if is_not_secure and cookie["secure"]:
                 continue
 
-            filtered[name] = cookie.value
+            # It's critical we use the Morsel so the coded_value
+            # (based on cookie version) is preserved
+            mrsl_val = cookie.get(cookie.key, Morsel())
+            mrsl_val.set(cookie.key, cookie.value, cookie.coded_value)
+            filtered[name] = mrsl_val
 
         return filtered
 
@@ -231,13 +244,13 @@ class CookieJar(AbstractCookieJar):
             return
 
         found_time = False
-        found_day_of_month = False
+        found_day = False
         found_month = False
         found_year = False
 
         hour = minute = second = 0
-        day_of_month = 0
-        month = ""
+        day = 0
+        month = 0
         year = 0
 
         for token_match in cls.DATE_TOKENS_RE.finditer(date_str):
@@ -252,18 +265,18 @@ class CookieJar(AbstractCookieJar):
                         int(s) for s in time_match.groups()]
                     continue
 
-            if not found_day_of_month:
-                day_of_month_match = cls.DATE_DAY_OF_MONTH_RE.match(token)
-                if day_of_month_match:
-                    found_day_of_month = True
-                    day_of_month = int(day_of_month_match.group())
+            if not found_day:
+                day_match = cls.DATE_DAY_OF_MONTH_RE.match(token)
+                if day_match:
+                    found_day = True
+                    day = int(day_match.group())
                     continue
 
             if not found_month:
                 month_match = cls.DATE_MONTH_RE.match(token)
                 if month_match:
                     found_month = True
-                    month = month_match.group()
+                    month = month_match.lastindex
                     continue
 
             if not found_year:
@@ -277,18 +290,15 @@ class CookieJar(AbstractCookieJar):
         elif 0 <= year <= 69:
             year += 2000
 
-        if False in (found_day_of_month, found_month, found_year, found_time):
+        if False in (found_day, found_month, found_year, found_time):
             return
 
-        if not 1 <= day_of_month <= 31:
+        if not 1 <= day <= 31:
             return
 
         if year < 1601 or hour > 23 or minute > 59 or second > 59:
             return
 
-        dt = datetime.datetime.strptime(
-            "%s %d %d:%d:%d %d" % (
-                month, day_of_month, hour, minute, second, year
-            ), "%b %d %H:%M:%S %Y")
-
-        return dt.replace(tzinfo=datetime.timezone.utc)
+        return datetime.datetime(year, month, day,
+                                 hour, minute, second,
+                                 tzinfo=datetime.timezone.utc)
