@@ -6,14 +6,17 @@ import binascii
 import cgi
 import datetime
 import functools
+import heapq
 import io
 import os
 import re
+import sys
+import time
 import warnings
 from collections import MutableSequence, namedtuple
 from functools import total_ordering
 from pathlib import Path
-from time import gmtime, time
+from time import gmtime
 from urllib.parse import urlencode
 
 from async_timeout import timeout
@@ -27,6 +30,12 @@ except ImportError:
     ensure_future = asyncio.async
 
 
+if sys.version_info >= (3, 4, 3):
+    from http.cookies import SimpleCookie  # noqa
+else:
+    from .backport_cookies import SimpleCookie  # noqa
+
+
 __all__ = ('BasicAuth', 'create_future', 'FormData', 'parse_mimetype',
            'Timeout', 'ensure_future')
 
@@ -96,11 +105,12 @@ class FormData:
     """Helper class for multipart/form-data and
     application/x-www-form-urlencoded body generation."""
 
-    def __init__(self, fields=()):
+    def __init__(self, fields=(), quote_fields=True):
         from . import multipart
         self._writer = multipart.MultipartWriter('form-data')
         self._fields = []
         self._is_multipart = False
+        self._quote_fields = quote_fields
 
         if isinstance(fields, dict):
             fields = list(fields.items())
@@ -191,7 +201,9 @@ class FormData:
         for dispparams, headers, value in self._fields:
             part = self._writer.append(value, headers)
             if dispparams:
-                part.set_content_disposition('form-data', **dispparams)
+                part.set_content_disposition(
+                    'form-data', quote_fields=self._quote_fields, **dispparams
+                )
                 # FIXME cgi.FieldStorage doesn't likes body parts with
                 # Content-Length which were sent via chunked transfer encoding
                 part.headers.pop(hdrs.CONTENT_LENGTH, None)
@@ -487,17 +499,17 @@ class reify:
         raise AttributeError("reified property is read-only")
 
 
-_ipv4_pattern = ('^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}'
-                 '(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$')
+_ipv4_pattern = (r'^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}'
+                 r'(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$')
 _ipv6_pattern = (
-    '^(?:(?:(?:[A-F0-9]{1,4}:){6}|(?=(?:[A-F0-9]{0,4}:){0,6}'
-    '(?:[0-9]{1,3}\.){3}[0-9]{1,3}$)(([0-9A-F]{1,4}:){0,5}|:)'
-    '((:[0-9A-F]{1,4}){1,5}:|:)|::(?:[A-F0-9]{1,4}:){5})'
-    '(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}'
-    '(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])|(?:[A-F0-9]{1,4}:){7}'
-    '[A-F0-9]{1,4}|(?=(?:[A-F0-9]{0,4}:){0,7}[A-F0-9]{0,4}$)'
-    '(([0-9A-F]{1,4}:){1,7}|:)((:[0-9A-F]{1,4}){1,7}|:)|(?:[A-F0-9]{1,4}:){7}'
-    ':|:(:[A-F0-9]{1,4}){7})$')
+    r'^(?:(?:(?:[A-F0-9]{1,4}:){6}|(?=(?:[A-F0-9]{0,4}:){0,6}'
+    r'(?:[0-9]{1,3}\.){3}[0-9]{1,3}$)(([0-9A-F]{1,4}:){0,5}|:)'
+    r'((:[0-9A-F]{1,4}){1,5}:|:)|::(?:[A-F0-9]{1,4}:){5})'
+    r'(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\.){3}'
+    r'(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])|(?:[A-F0-9]{1,4}:){7}'
+    r'[A-F0-9]{1,4}|(?=(?:[A-F0-9]{0,4}:){0,7}[A-F0-9]{0,4}$)'
+    r'(([0-9A-F]{1,4}:){1,7}|:)((:[0-9A-F]{1,4}){1,7}|:)|(?:[A-F0-9]{1,4}:){7}'
+    r':|:(:[A-F0-9]{1,4}){7})$')
 _ipv4_regex = re.compile(_ipv4_pattern)
 _ipv6_regex = re.compile(_ipv6_pattern, flags=re.IGNORECASE)
 _ipv4_regexb = re.compile(_ipv4_pattern.encode('ascii'))
@@ -583,29 +595,63 @@ class FrozenList(MutableSequence):
         self._items.insert(pos, item)
 
 
+class TimerHandle(asyncio.TimerHandle):
+
+    def cancel(self):
+        asyncio.Handle.cancel(self)
+
+
 class TimeService:
-    def __init__(self, loop):
+
+    def __init__(self, loop, *, interval=1.0):
         self._loop = loop
-        self._time = time()
+        self._interval = interval
+        self._time = time.time()
+        self._loop_time = loop.time()
+        self._count = 0
         self._strtime = None
-        self._count = 0
-        self._cb = loop.call_later(1, self._on_cb)
+        self._cb = loop.call_at(self._loop_time + self._interval, self._on_cb)
+        self._scheduled = []
+
+    def close(self):
+        if self._cb:
+            self._cb.cancel()
+
+        # cancel all scheduled handles
+        for handle in self._scheduled:
+            handle.cancel()
 
-    def stop(self):
-        self._cb.cancel()
         self._cb = None
+        self._scheduled = []
         self._loop = None
 
-    def _on_cb(self):
-        self._count += 1
-        if self._count >= 10*60:
+    def _on_cb(self, reset_count=10*60):
+        self._loop_time = self._loop.time()
+
+        if self._count >= reset_count:
             # reset timer every 10 minutes
             self._count = 0
-            self._time = time()
+            self._time = time.time()
         else:
-            self._time += 1
+            self._time += self._interval
+
+        # Handle 'later' callbacks that are ready.
+        ready = []
+        end_time = self._loop_time
+        while self._scheduled:
+            handle = self._scheduled[0]
+            if handle._when >= end_time:
+                break
+            handle = heapq.heappop(self._scheduled)
+            ready.append(handle)
+
+        for handle in ready:
+            if not handle._cancelled:
+                handle._run()
+
         self._strtime = None
-        self._cb = self._loop.call_later(1, self._on_cb)
+        self._cb = self._loop.call_at(
+            self._loop_time + self._interval, self._on_cb)
 
     def _format_date_time(self):
         # Weekday and month names for HTTP date/time formatting;
@@ -630,6 +676,95 @@ class TimeService:
             self._strtime = s = self._format_date_time()
         return self._strtime
 
+    def loop_time(self):
+        return self._loop_time
+
+    def call_later(self, delay, callback, *args):
+        """Arrange for a callback to be called at a given time.
+
+        Return a Handle: an opaque object with a cancel() method that
+        can be used to cancel the call.
+
+        The delay can be an int or float, expressed in seconds.  It is
+        always relative to the current time.
+
+        Any positional arguments after the callback will be passed to
+        the callback when it is called.
+
+        Time resolution is aproximatly one second.
+        """
+        return self._call_at(self._loop_time + delay, callback, *args)
+
+    def _call_at(self, when, callback, *args):
+        """Like call_later(), but uses an absolute time.
+
+        Absolute time corresponds to the loop's time() method.
+        """
+        timer = TimerHandle(when, callback, args, self._loop)
+        heapq.heappush(self._scheduled, timer)
+        return timer
+
+    def timeout(self, timeout):
+        """low resolution timeout context manager.
+
+        timeout - value in seconds or None to disable timeout logic
+        """
+        if timeout:
+            when = self._loop_time + timeout
+            ctx = _TimeServiceTimeoutContext(when, self._loop)
+            heapq.heappush(self._scheduled, ctx)
+        else:
+            ctx = _TimeServiceTimeoutNoop()
+
+        return ctx
+
+
+class _TimeServiceTimeoutNoop:
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        return False
+
+
+class _TimeServiceTimeoutContext(TimerHandle):
+    """ Low resolution timeout context manager """
+
+    def __init__(self, when, loop):
+        super().__init__(when, self.cancel, (), loop)
+
+        self._tasks = []
+        self._cancelled = False
+
+    def __enter__(self):
+        task = asyncio.Task.current_task(loop=self._loop)
+        if task is None:
+            raise RuntimeError('Timeout context manager should be used '
+                               'inside a task')
+
+        if self._cancelled:
+            task.cancel()
+            raise asyncio.TimeoutError from None
+
+        self._tasks.append(task)
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        if self._tasks:
+            self._tasks.pop()
+
+        if exc_type is asyncio.CancelledError and self._cancelled:
+            raise asyncio.TimeoutError from None
+
+    def cancel(self):
+        if not self._cancelled:
+            for task in self._tasks:
+                task.cancel()
+
+            self._tasks = []
+            self._cancelled = True
+
 
 class HeadersMixin:
 
@@ -670,8 +805,3 @@ class HeadersMixin:
             return None
         else:
             return int(l)
-
-
-def check_loop(loop):
-    if loop is None:
-        loop = asyncio.get_event_loop()
