@@ -1,7 +1,6 @@
 import asyncio
 import contextlib
 import gc
-import http.cookies
 import re
 import types
 from unittest import mock
@@ -14,6 +13,7 @@ import aiohttp
 from aiohttp import web
 from aiohttp.client import ClientSession
 from aiohttp.connector import BaseConnector, TCPConnector
+from aiohttp.helpers import SimpleCookie
 
 
 @pytest.fixture
@@ -273,9 +273,9 @@ def test_connector_loop(loop):
         stack.enter_context(contextlib.closing(another_loop))
         connector = TCPConnector(loop=another_loop)
         stack.enter_context(contextlib.closing(connector))
-        with pytest.raises(ValueError) as ctx:
+        with pytest.raises(RuntimeError) as ctx:
             ClientSession(connector=connector, loop=loop)
-        assert re.match("loop argument must agree with connector",
+        assert re.match("Session and connector has to use same event loop",
                         str(ctx.value))
 
 
@@ -371,6 +371,7 @@ def test_request_ctx_manager_props(loop):
         assert isinstance(ctx_mgr.gi_frame, types.FrameType)
         assert not ctx_mgr.gi_running
         assert isinstance(ctx_mgr.gi_code, types.CodeType)
+        yield from asyncio.sleep(0.1, loop=loop)
 
 
 @asyncio.coroutine
@@ -409,7 +410,7 @@ def test_cookie_jar_usage(loop, test_client):
     # Updating the cookie jar with the response cookies
     assert jar.update_cookies.called
     resp_cookies = jar.update_cookies.call_args[0][0]
-    assert isinstance(resp_cookies, http.cookies.SimpleCookie)
+    assert isinstance(resp_cookies, SimpleCookie)
     assert "response" in resp_cookies
     assert resp_cookies["response"].value == "resp_value"
 
@@ -422,6 +423,7 @@ def test_session_default_version(loop):
 def test_session_loop(loop):
     session = aiohttp.ClientSession(loop=loop)
     assert session.loop is loop
+    session.close()
 
 
 def test_proxy_str(session, params):
@@ -435,3 +437,16 @@ def test_proxy_str(session, params):
                                            allow_redirects=True,
                                            proxy='http://proxy.com',
                                            **params)]
+
+
+def test_client_session_implicit_loop_warn():
+    loop = asyncio.new_event_loop()
+    asyncio.set_event_loop(loop)
+
+    with pytest.warns(ResourceWarning):
+        session = aiohttp.ClientSession()
+        assert session._loop is loop
+        session.close()
+
+    asyncio.set_event_loop(None)
+    loop.close()
