@@ -6,10 +6,10 @@ from unittest import mock
 
 import pytest
 from multidict import MultiDict
+from yarl import URL
 
 from aiohttp import FormData, multipart, web
 from aiohttp.protocol import HttpVersion, HttpVersion10, HttpVersion11
-from aiohttp.streams import EOF_MARKER
 
 try:
     import ssl
@@ -52,7 +52,7 @@ def test_handler_returns_not_response(loop, test_server, test_client):
     resp = yield from client.get('/')
     assert 500 == resp.status
 
-    logger.exception.assert_called_with("Error handling request")
+    assert logger.exception.called
 
 
 @asyncio.coroutine
@@ -334,7 +334,7 @@ def test_release_post_data(loop, test_client):
     def handler(request):
         yield from request.release()
         chunk = yield from request.content.readany()
-        assert chunk is EOF_MARKER
+        assert chunk == b''
         return web.Response()
 
     app = web.Application(loop=loop)
@@ -384,7 +384,7 @@ def test_post_form_with_duplicate_keys(loop, test_client):
 
 def test_repr_for_application(loop):
     app = web.Application(loop=loop)
-    assert "<Application>" == repr(app)
+    assert "<Application 0x{:x}>".format(id(app)) == repr(app)
 
 
 @asyncio.coroutine
@@ -622,7 +622,7 @@ def test_http10_keep_alive_with_headers(loop, test_client):
 def test_upload_file(loop, test_client):
 
     here = pathlib.Path(__file__).parent
-    fname = here / 'software_development_in_picture.jpg'
+    fname = here / 'aiohttp.png'
     with fname.open('rb') as f:
         data = f.read()
 
@@ -644,7 +644,7 @@ def test_upload_file(loop, test_client):
 @asyncio.coroutine
 def test_upload_file_object(loop, test_client):
     here = pathlib.Path(__file__).parent
-    fname = here / 'software_development_in_picture.jpg'
+    fname = here / 'aiohttp.png'
     with fname.open('rb') as f:
         data = f.read()
 
@@ -763,14 +763,36 @@ def test_get_with_empty_arg_with_equal(loop, test_client):
     assert 200 == resp.status
 
 
-@pytest.mark.xfail  # and had never worked
 @asyncio.coroutine
-def test_response_with_precompressed_body(loop, test_client):
+def test_response_with_precompressed_body_gzip(loop, test_client):
+
     @asyncio.coroutine
     def handler(request):
         headers = {'Content-Encoding': 'gzip'}
-        deflated_data = zlib.compress(b'mydata')
-        return web.Response(body=deflated_data, headers=headers)
+        zcomp = zlib.compressobj(wbits=16 + zlib.MAX_WBITS)
+        data = zcomp.compress(b'mydata') + zcomp.flush()
+        return web.Response(body=data, headers=headers)
+
+    app = web.Application(loop=loop)
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    data = yield from resp.read()
+    assert b'mydata' == data
+    assert resp.headers.get('Content-Encoding') == 'gzip'
+
+
+@asyncio.coroutine
+def test_response_with_precompressed_body_deflate(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        headers = {'Content-Encoding': 'deflate'}
+        zcomp = zlib.compressobj(wbits=-zlib.MAX_WBITS)
+        data = zcomp.compress(b'mydata') + zcomp.flush()
+        return web.Response(body=data, headers=headers)
 
     app = web.Application(loop=loop)
     app.router.add_get('/', handler)
@@ -825,16 +847,381 @@ def test_requests_count(loop, test_client):
     app = web.Application(loop=loop)
     app.router.add_get('/', handler)
     client = yield from test_client(app)
-    assert client.handler.requests_count == 0
+    assert client.server.handler.requests_count == 0
 
     resp = yield from client.get('/')
     assert 200 == resp.status
-    assert client.handler.requests_count == 1
+    assert client.server.handler.requests_count == 1
 
     resp = yield from client.get('/')
     assert 200 == resp.status
-    assert client.handler.requests_count == 2
+    assert client.server.handler.requests_count == 2
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    assert client.server.handler.requests_count == 3
+
+
+@asyncio.coroutine
+def test_redirect_url(loop, test_client):
+
+    @asyncio.coroutine
+    def redirector(request):
+        raise web.HTTPFound(location=URL('/redirected'))
+
+    @asyncio.coroutine
+    def redirected(request):
+        return web.Response()
+
+    app = web.Application(loop=loop)
+    app.router.add_get('/redirector', redirector)
+    app.router.add_get('/redirected', redirected)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/redirector')
+    assert resp.status == 200
+
+
+@asyncio.coroutine
+def test_simple_subapp(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(text="OK")
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    app.add_subapp('/path', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/path/to')
+    assert resp.status == 200
+    txt = yield from resp.text()
+    assert 'OK' == txt
+
+
+@asyncio.coroutine
+def test_subapp_reverse_url(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPMovedPermanently(
+            location=subapp.router['name'].url_for())
+
+    @asyncio.coroutine
+    def handler2(request):
+        return web.Response(text="OK")
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    subapp.router.add_get('/final', handler2, name='name')
+    app.add_subapp('/path', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/path/to')
+    assert resp.status == 200
+    txt = yield from resp.text()
+    assert 'OK' == txt
+    assert resp.url_obj.path == '/path/final'
+
+
+@asyncio.coroutine
+def test_subapp_reverse_variable_url(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPMovedPermanently(
+            location=subapp.router['name'].url_for(part='final'))
+
+    @asyncio.coroutine
+    def handler2(request):
+        return web.Response(text="OK")
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    subapp.router.add_get('/{part}', handler2, name='name')
+    app.add_subapp('/path', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/path/to')
+    assert resp.status == 200
+    txt = yield from resp.text()
+    assert 'OK' == txt
+    assert resp.url_obj.path == '/path/final'
+
+
+@asyncio.coroutine
+def test_subapp_reverse_static_url(loop, test_client):
+    fname = 'aiohttp.png'
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPMovedPermanently(
+            location=subapp.router['name'].url_for(filename=fname))
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    here = pathlib.Path(__file__).parent
+    subapp.router.add_static('/static', here, name='name')
+    app.add_subapp('/path', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/path/to')
+    assert resp.url_obj.path == '/path/static/' + fname
+    assert resp.status == 200
+    body = yield from resp.read()
+    with (here / fname).open('rb') as f:
+        assert body == f.read()
+
+
+@asyncio.coroutine
+def test_subapp_app(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        assert request.app is subapp
+        return web.HTTPOk(text='OK')
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    app.add_subapp('/path/', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/path/to')
+    assert resp.status == 200
+    txt = yield from resp.text()
+    assert 'OK' == txt
+
+
+@asyncio.coroutine
+def test_subapp_not_found(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPOk(text='OK')
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    app.add_subapp('/path/', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/path/other')
+    assert resp.status == 404
+
+
+@asyncio.coroutine
+def test_subapp_not_found2(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPOk(text='OK')
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    app.add_subapp('/path/', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/invalid/other')
+    assert resp.status == 404
+
+
+@asyncio.coroutine
+def test_subapp_not_allowed(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPOk(text='OK')
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    app.add_subapp('/path/', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.post('/path/to')
+    assert resp.status == 405
+    assert resp.headers['Allow'] == 'GET'
+
+
+@asyncio.coroutine
+def test_subapp_cannot_add_app_in_handler(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        request.match_info.add_app(app)
+        return web.HTTPOk(text='OK')
+
+    app = web.Application(loop=loop)
+    subapp = web.Application(loop=loop)
+    subapp.router.add_get('/to', handler)
+    app.add_subapp('/path/', subapp)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/path/to')
+    assert resp.status == 500
+
+
+@asyncio.coroutine
+def test_subapp_middlewares(loop, test_client):
+    order = []
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPOk(text='OK')
+
+    @asyncio.coroutine
+    def middleware_factory(app, handler):
+
+        @asyncio.coroutine
+        def middleware(request):
+            order.append((1, app))
+            resp = yield from handler(request)
+            assert 200 == resp.status
+            order.append((2, app))
+            return resp
+        return middleware
+
+    app = web.Application(loop=loop, middlewares=[middleware_factory])
+    subapp1 = web.Application(loop=loop, middlewares=[middleware_factory])
+    subapp2 = web.Application(loop=loop, middlewares=[middleware_factory])
+    subapp2.router.add_get('/to', handler)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/a/b/to')
+    assert resp.status == 200
+    assert [(1, app), (1, subapp1), (1, subapp2),
+            (2, subapp2), (2, subapp1), (2, app)] == order
+
+
+@asyncio.coroutine
+def test_subapp_on_response_prepare(loop, test_client):
+    order = []
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.HTTPOk(text='OK')
+
+    def make_signal(app):
+
+        @asyncio.coroutine
+        def on_response(request, response):
+            order.append(app)
+
+        return on_response
+
+    app = web.Application(loop=loop)
+    app.on_response_prepare.append(make_signal(app))
+    subapp1 = web.Application(loop=loop)
+    subapp1.on_response_prepare.append(make_signal(subapp1))
+    subapp2 = web.Application(loop=loop)
+    subapp2.on_response_prepare.append(make_signal(subapp2))
+    subapp2.router.add_get('/to', handler)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
+
+    client = yield from test_client(app)
+    resp = yield from client.get('/a/b/to')
+    assert resp.status == 200
+    assert [app, subapp1, subapp2] == order
+
+
+@asyncio.coroutine
+def test_subapp_on_startup(loop, test_server):
+    order = []
+
+    @asyncio.coroutine
+    def on_signal(app):
+        order.append(app)
+
+    app = web.Application(loop=loop)
+    app.on_startup.append(on_signal)
+    subapp1 = web.Application(loop=loop)
+    subapp1.on_startup.append(on_signal)
+    subapp2 = web.Application(loop=loop)
+    subapp2.on_startup.append(on_signal)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
+
+    yield from test_server(app)
+
+    assert [app, subapp1, subapp2] == order
+
+
+@asyncio.coroutine
+def test_subapp_on_shutdown(loop, test_server):
+    order = []
+
+    def on_signal(app):
+        order.append(app)
+
+    app = web.Application(loop=loop)
+    app.on_shutdown.append(on_signal)
+    subapp1 = web.Application(loop=loop)
+    subapp1.on_shutdown.append(on_signal)
+    subapp2 = web.Application(loop=loop)
+    subapp2.on_shutdown.append(on_signal)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
+
+    server = yield from test_server(app)
+    yield from server.close()
+
+    assert [app, subapp1, subapp2] == order
+
+
+@asyncio.coroutine
+def test_subapp_on_cleanup(loop, test_server):
+    order = []
+
+    @asyncio.coroutine
+    def on_signal(app):
+        order.append(app)
+
+    app = web.Application(loop=loop)
+    app.on_cleanup.append(on_signal)
+    subapp1 = web.Application(loop=loop)
+    subapp1.on_cleanup.append(on_signal)
+    subapp2 = web.Application(loop=loop)
+    subapp2.on_cleanup.append(on_signal)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
+
+    server = yield from test_server(app)
+    yield from server.close()
+
+    assert [app, subapp1, subapp2] == order
+
+
+@asyncio.coroutine
+def test_custom_date_header(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(headers={'Date': 'Sun, 30 Oct 2016 03:13:52 GMT'})
+
+    app = web.Application(loop=loop)
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    assert resp.headers['Date'] == 'Sun, 30 Oct 2016 03:13:52 GMT'
+
+
+@asyncio.coroutine
+def test_response_prepared_with_clone(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        cloned = request.clone()
+        resp = web.StreamResponse()
+        yield from resp.prepare(cloned)
+        return resp
+
+    app = web.Application(loop=loop)
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
 
     resp = yield from client.get('/')
     assert 200 == resp.status
-    assert client.handler.requests_count == 3
