@@ -1,10 +1,13 @@
 import asyncio
+from collections import MutableMapping
 from unittest import mock
 
 import pytest
 from multidict import CIMultiDict, MultiDict
+from yarl import URL
 
 from aiohttp.protocol import HttpVersion
+from aiohttp.streams import StreamReader
 from aiohttp.test_utils import make_mocked_request
 
 
@@ -52,8 +55,9 @@ def test_ctor(make_request):
 
 
 def test_doubleslashes(make_request):
-    req = make_request('GET', '//foo/')
-    assert '//foo/' == req.path
+    # NB: //foo/bar is an absolute URL with foo netloc and /bar path
+    req = make_request('GET', '/bar//foo/')
+    assert '/bar//foo/' == req.path
 
 
 def test_POST(make_request):
@@ -105,6 +109,11 @@ def test_non_ascii_path(make_request):
     assert '/путь' == req.path
 
 
+def test_non_ascii_raw_path(make_request):
+    req = make_request('GET', '/путь')
+    assert '/%D0%BF%D1%83%D1%82%D1%8C' == req.raw_path
+
+
 def test_content_length(make_request):
     req = make_request('Get', '/',
                        CIMultiDict([('CONTENT-LENGTH', '123')]))
@@ -178,23 +187,36 @@ def test_request_cookie__set_item(make_request):
 
 def test_match_info(make_request):
     req = make_request('GET', '/')
-    assert req.match_info is None
-    match = {'a': 'b'}
-    req._match_info = match
-    assert match is req.match_info
+    assert req._match_info is req.match_info
 
 
-def test_request_is_dict(make_request):
+def test_request_is_mutable_mapping(make_request):
     req = make_request('GET', '/')
-    assert isinstance(req, dict)
+    assert isinstance(req, MutableMapping)
     req['key'] = 'value'
     assert 'value' == req['key']
 
 
-def test_copy(make_request):
+def test_request_delitem(make_request):
     req = make_request('GET', '/')
-    with pytest.raises(NotImplementedError):
-        req.copy()
+    req['key'] = 'value'
+    assert 'value' == req['key']
+    del req['key']
+    assert 'key' not in req
+
+
+def test_request_len(make_request):
+    req = make_request('GET', '/')
+    assert len(req) == 0
+    req['key'] = 'value'
+    assert len(req) == 1
+
+
+def test_request_iter(make_request):
+    req = make_request('GET', '/')
+    req['key'] = 'value'
+    req['key2'] = 'value2'
+    assert set(req) == {'key', 'key2'}
 
 
 def test___repr__(make_request):
@@ -235,3 +257,64 @@ def test_raw_headers(make_request):
     req = make_request('GET', '/',
                        headers=CIMultiDict({'X-HEADER': 'aaa'}))
     assert req.raw_headers == ((b'X-Header', b'aaa'),)
+
+
+def test_rel_url(make_request):
+    req = make_request('GET', '/path')
+    assert URL('/path') == req.rel_url
+
+
+def test_url_url(make_request):
+    req = make_request('GET', '/path', headers={'HOST': 'example.com'})
+    assert URL('http://example.com/path') == req.url
+
+
+def test_clone():
+    req = make_mocked_request('GET', '/path')
+    req2 = req.clone()
+    assert req2.method == 'GET'
+    assert req2.rel_url == URL('/path')
+
+
+def test_clone_method():
+    req = make_mocked_request('GET', '/path')
+    req2 = req.clone(method='POST')
+    assert req2.method == 'POST'
+    assert req2.rel_url == URL('/path')
+
+
+def test_clone_rel_url():
+    req = make_mocked_request('GET', '/path')
+    req2 = req.clone(rel_url=URL('/path2'))
+    assert req2.rel_url == URL('/path2')
+
+
+def test_clone_rel_url_str():
+    req = make_mocked_request('GET', '/path')
+    req2 = req.clone(rel_url='/path2')
+    assert req2.rel_url == URL('/path2')
+
+
+def test_clone_headers():
+    req = make_mocked_request('GET', '/path', headers={'A': 'B'})
+    req2 = req.clone(headers=CIMultiDict({'B': 'C'}))
+    assert req2.headers == CIMultiDict({'B': 'C'})
+    assert req2.raw_headers == ((b'B', b'C'),)
+
+
+def test_clone_headers_dict():
+    req = make_mocked_request('GET', '/path', headers={'A': 'B'})
+    req2 = req.clone(headers={'B': 'C'})
+    assert req2.headers == CIMultiDict({'B': 'C'})
+    assert req2.raw_headers == ((b'B', b'C'),)
+
+
+@asyncio.coroutine
+def test_cannot_clone_after_read(loop):
+    payload = StreamReader(loop=loop)
+    payload.feed_data(b'data')
+    payload.feed_eof()
+    req = make_mocked_request('GET', '/path', payload=payload)
+    yield from req.read()
+    with pytest.raises(RuntimeError):
+        req.clone()
