@@ -1,6 +1,7 @@
 """Tests for aiohttp/worker.py"""
 import asyncio
 import pathlib
+import socket
 import ssl
 from unittest import mock
 
@@ -76,16 +77,57 @@ def test_run(worker, loop):
     assert loop.is_closed()
 
 
+def test_run_wsgi(worker, loop):
+    worker.wsgi = lambda env, start_resp: start_resp()
+
+    worker.loop = loop
+    worker._run = mock.Mock(
+        wraps=asyncio.coroutine(lambda: None))
+    with pytest.raises(SystemExit):
+        worker.run()
+    assert worker._run.called
+    assert loop.is_closed()
+
+
 def test_handle_quit(worker):
-    worker.handle_quit(object(), object())
-    assert not worker.alive
-    assert worker.exit_code == 0
+    with mock.patch('aiohttp.worker.ensure_future') as m_ensure_future:
+        worker.loop = mock.Mock()
+        worker.handle_quit(object(), object())
+        assert not worker.alive
+        assert worker.exit_code == 0
+        assert m_ensure_future.called
+        worker.loop.call_later.asset_called_with(
+            0.1, worker._notify_waiter_done)
 
 
 def test_handle_abort(worker):
-    worker.handle_abort(object(), object())
-    assert not worker.alive
-    assert worker.exit_code == 1
+    with mock.patch('aiohttp.worker.sys') as m_sys:
+        worker.handle_abort(object(), object())
+        assert not worker.alive
+        assert worker.exit_code == 1
+        m_sys.exit.assert_called_with(1)
+
+
+def test__wait_next_notify(worker):
+    worker.loop = mock.Mock()
+    worker._notify_waiter_done = mock.Mock()
+    fut = worker._wait_next_notify()
+
+    assert worker._notify_waiter == fut
+    worker.loop.call_later.assert_called_with(1.0, worker._notify_waiter_done)
+
+
+def test__notify_waiter_done(worker):
+    worker._notify_waiter = None
+    worker._notify_waiter_done()
+    assert worker._notify_waiter is None
+
+    waiter = worker._notify_waiter = mock.Mock()
+    worker._notify_waiter.done.return_value = False
+    worker._notify_waiter_done()
+
+    assert worker._notify_waiter is None
+    waiter.set_result.assert_called_with(True)
 
 
 def test_init_signals(worker):
@@ -107,6 +149,21 @@ def test_make_handler(worker, mocker):
     assert worker._get_valid_log_format.called
 
 
+def test_make_handler_wsgi(worker, mocker):
+    worker.wsgi = lambda env, start_resp: start_resp()
+    worker.loop = mock.Mock()
+    worker.loop.time.return_value = 1477797232
+    worker.log = mock.Mock()
+    worker.cfg = mock.Mock()
+    worker.cfg.access_log_format = ACCEPTABLE_LOG_FORMAT
+    mocker.spy(worker, '_get_valid_log_format')
+
+    f = worker.make_handler(worker.wsgi)
+    assert isinstance(f, base_worker.WSGIServer)
+    assert isinstance(f(), base_worker.WSGIServerHttpProtocol)
+    assert worker._get_valid_log_format.called
+
+
 @pytest.mark.parametrize('source,result', [
     (ACCEPTABLE_LOG_FORMAT, ACCEPTABLE_LOG_FORMAT),
     (AsyncioWorker.DEFAULT_GUNICORN_LOG_FORMAT,
@@ -157,6 +214,44 @@ def test__run_ok(worker, loop):
     assert ctx is ssl_context
 
 
+@pytest.mark.skipif(not hasattr(socket, 'AF_UNIX'),
+                    reason="UNIX sockets are not supported")
+@asyncio.coroutine
+def test__run_ok_unix_socket(worker, loop):
+    worker.ppid = 1
+    worker.alive = True
+    worker.servers = {}
+    sock = mock.Mock()
+    sock.cfg_addr = ('/path/to')
+    sock.family = socket.AF_UNIX
+    worker.sockets = [sock]
+    worker.wsgi = mock.Mock()
+    worker.close = make_mocked_coro(None)
+    worker.log = mock.Mock()
+    worker.loop = loop
+    loop.create_unix_server = make_mocked_coro(sock)
+    worker.wsgi.make_handler.return_value.requests_count = 1
+    worker.cfg.max_requests = 100
+    worker.cfg.is_ssl = True
+    worker.cfg.access_log_format = ACCEPTABLE_LOG_FORMAT
+
+    ssl_context = mock.Mock()
+    with mock.patch('ssl.SSLContext', return_value=ssl_context):
+        with mock.patch('aiohttp.worker.asyncio') as m_asyncio:
+            m_asyncio.sleep = mock.Mock(
+                wraps=asyncio.coroutine(lambda *a, **kw: None))
+            yield from worker._run()
+
+    worker.notify.assert_called_with()
+    worker.log.info.assert_called_with("Parent changed, shutting down: %s",
+                                       worker)
+
+    args, kwargs = loop.create_unix_server.call_args
+    assert 'ssl' in kwargs
+    ctx = kwargs['ssl']
+    assert ctx is ssl_context
+
+
 @asyncio.coroutine
 def test__run_exc(worker, loop):
     with mock.patch('aiohttp.worker.os') as m_os:
@@ -166,6 +261,7 @@ def test__run_exc(worker, loop):
         handler = mock.Mock()
         handler.requests_count = 0
         worker.servers = {mock.Mock(): handler}
+        worker._wait_next_notify = mock.Mock()
         worker.ppid = 1
         worker.alive = True
         worker.sockets = []
@@ -184,7 +280,7 @@ def test__run_exc(worker, loop):
 
             yield from worker._run()
 
-        m_sleep.assert_called_with(1.0, loop=loop)
+        assert worker._wait_next_notify.called
         worker.close.assert_called_with()
 
 
@@ -199,8 +295,8 @@ def test_close(worker, loop):
     app = worker.wsgi = mock.Mock()
     app.cleanup = make_mocked_coro(None)
     handler.connections = [object()]
-    handler.finish_connections.return_value = helpers.create_future(loop)
-    handler.finish_connections.return_value.set_result(1)
+    handler.shutdown.return_value = helpers.create_future(loop)
+    handler.shutdown.return_value.set_result(1)
 
     app.shutdown.return_value = helpers.create_future(loop)
     app.shutdown.return_value.set_result(None)
@@ -208,7 +304,28 @@ def test_close(worker, loop):
     yield from worker.close()
     app.shutdown.assert_called_with()
     app.cleanup.assert_called_with()
-    handler.finish_connections.assert_called_with(timeout=95.0)
+    handler.shutdown.assert_called_with(timeout=95.0)
+    srv.close.assert_called_with()
+    assert worker.servers is None
+
+    yield from worker.close()
+
+
+@asyncio.coroutine
+def test_close_wsgi(worker, loop):
+    srv = mock.Mock()
+    srv.wait_closed = make_mocked_coro(None)
+    handler = mock.Mock()
+    worker.servers = {srv: handler}
+    worker.log = mock.Mock()
+    worker.loop = loop
+    worker.wsgi = lambda env, start_resp: start_resp()
+    handler.connections = [object()]
+    handler.shutdown.return_value = helpers.create_future(loop)
+    handler.shutdown.return_value.set_result(1)
+
+    yield from worker.close()
+    handler.shutdown.assert_called_with(timeout=95.0)
     srv.close.assert_called_with()
     assert worker.servers is None
 
